#!/usr/bin/env node

var args = require('yargs'),
    M = require('mstring'),
    pkg = require('../package.json'),
    EnvController = require('../lib/controllers/EnvironmentController'),
    AppController = require('../lib/controllers/AppController'),
    prompt = require('prompt'),
    path = require('path');


var usage = M(function() {

    /***

     Commands:
       $0 new-app             <app-name>        [-t --title --description]
       $0 publish-app                           [-e -n -u -p]
       $0 list-apps                             [-e -n -u -p]
       $0 list-profile                         [-e -n -u -p]
       $0 set-profile         <profile-name>    []
     ***/

});

// TODO: make this better!
var argv = args
    .usage(usage)
    .options('t', {alias:'type', describe: 'The project type (custom | angular | app-generator).'})
    .options('title', {describe: 'The display name of the app.'})
    .options('description', {describe: 'The description of the app.'})
    .options('e', {alias:'environment', describe: 'The environment to use. Not needed if you want to use the default environment.'})
    .options('n', {alias:'namespace',   describe: 'The namespace. Not needed if environment is specified.'})
    .options('u', {alias:'user',    describe: 'The username. Not needed if environment is specified.'})
    .options('p', {alias:'password',    describe: "The password. Not needed if environment is specified."})
    .options('s', {alias: 'select', describe: ''})
    .options('i', {alias: 'insert', describe: ''})
    .options('d', {alias: 'delete', describe: ''})
    .options('U', {alias: 'update', describe: ''})
    .argv;


//Just want the version
if(argv.version){
    console.log(pkg.version);
    return;
}

//Just want the help
if(argv.help){
    args.showHelp();
    return;
}

function AppTools(options) {

    this.cwd          = process.cwd();   // save this before loadProjectSync changes it
    this.helpers      = require('../lib/helpers')(argv);

    this.project      = this.helpers.loadProjectSync();
    this.createEnv    = this.helpers.createEnv;

    this.controllers  = {
        env: EnvController(this.project, options)
    }
}

AppTools.prototype = {

    // Parse out command
    dispatch: function(argv) {
        var action = (argv._ && argv._.length > 0) ? argv._.shift() : '';
        this.action = action.toLowerCase();

        switch(action){
            case '':
                this.helpers.fatal('No command was found.');
                break;
            case 'new-app':             // done
                this.newApp(argv);
                break;
            case 'publish-app':         // done
                this.publishApp(argv);
                break;
//            case 'add-profile':         // done
//                this.addProfile(argv);
//                break;
//            case 'remove-profile':      // done
//                this.removeProfile(argv);
//                break;
//            case 'change-profile-name':
//                this.changeProfileName(argv);
//                break;
            case 'view-profile':
                this.viewProfile(argv);
                break;
            case 'add-type-to-profile':
                this.addTypeToProfile(argv);
                break;
            case 'remove-type-from-profile':
                this.removeTypeFromProfile(argv);
                break;
            case 'list-profile':       // done
                this.listProfile(argv);
                break;
            case 'assign-profile-to-user':
                this.assignProfileToUser(argv);
                break;
            case 'list-apps':           // done => need to fix output
                this.listApps(argv);
                break;
            case 'create-app-metadata': // todo
                // TODO
                    // this is for apps not being hosted in cloud
                    // NOTE: need to add custom location (url) where app resides
                break;
            default:
                this.helpers.fatal('Unknown Command: ' + action + '.');
        }
    },

    // Commands
    newApp: function(argv) {    // Args: -t, --title, --description
                                    // need: -m, -e, -n, -u, -p for app generator

        // Parse inputs
        var name = argv._.shift(),
            type = argv.t || 'custom',
            model = argv.m;

        if (!name) {
            this.helpers.fatal('Missing project name.');
        }
        var pathToAppFolder = path.join(process.cwd(), 'apps', name);

        this.demandProjectOrExit();

        this.helpers.checkIfAppExists(pathToAppFolder);

        // Prompt for info
        var self = this;
        prompt.message = 'app-tools';
        prompt.override = argv;
        prompt.start();
        prompt.get([
            {
                name: "title",
                description: "display name of app (required)".cyan,
                required: true
            },
            {
                name: "description",
                description: "description of app (optional)".cyan,
                required: false
            }
        ], function(err, result) {

            if (err) {
                self.helpers.fatal('\nAborted app creation.');
            }

            self.helpers.createAppDirectory(pathToAppFolder);

            // Generate skeleton project
            type = type.toLowerCase();
            switch (type) {
                case 'custom':
                    self.helpers.createCustomApp(name, result.title, result.description, pathToAppFolder)
                        .then(function(msg) {
                            console.log(msg);
                        }, function(msg) {
                            self.helpers.fatal(msg);
                        });
                    break;
                case 'app-generator':

                    var credentials = self.helpers.validateCredentials(self.project, argv);

                    if (model) {
                        self.helpers.createAppGeneratorApp(model, name, result.title, result.description, pathToAppFolder, credentials);
                    } else {
                        self.helpers.fatal('Failed to create app for type: ' + model + '. Please include -m.');
                    }

                    break;
                case 'angular':
                    // TODO
                    self.helpers.createAngularSeedApp(name, result.title, result.description, pathToAppFolder);
                    break;
                default:
                    // in case user spells it wrong...
                    self.fatal('Unknown project type. Please specify either custom, angular, or app-generator.');
            }

        });

    },

    // ASSUMPTION: you have to be in the app directory to publish!
    publishApp: function(argv) {    // Possible Args: -e -n -u -p --dir (optional -> default to /www)
        // ENDPOINT: PUT or POST to /apps

        // Parse inputs
        var publishDirectory = argv.dir || 'www';   // TODO: take out choice - will only be /www

        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Validate
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'index.html');
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'package.json');
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'LICENSE');

        // Create app instance
        // NOTE: for this command only we will use the package.json from the publish directory
        this.app = this.helpers.loadPackageSync(path.join(pathToAppFolder, publishDirectory));
        this.controllers.app = AppController(this.app);

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Make sure all the required attributes are present before publishing
        var validApp = this.controllers.app.validateApp();
        if (!validApp.valid) {
            this.helpers.fatal(validApp.msg);
//            this.helpers.fatal('Missing required fields in package.json. Note: both name and version are required.');

            // TODO: prompt for missing required fields?
        }

        // Save overwrites to package.json
        this.app.save();

        // Create Zip
        var self = this;
        this.helpers.createAppZip(this.app.getName(), pathToAppFolder, publishDirectory, new Date().getTime())
            .then(function(obj) {

                var zipName = obj.zipName,
                    size    = obj.size;

                console.log(obj.msg);

                // Publish App
                return self.helpers.sendApp(zipName, size, pathToAppFolder, credentials, self.app);

            })
            .then(function(response) {

                console.log('got here');
                console.log('message', response.message);

                if (response.type === 'POST') {
                    // Assign Profile to publisher - ONLY ON POST!
                    return self.helpers.assignProfile(self.app.getProfile().name, credentials.username, credentials);
                }

            }, function(msg) {

                self.helpers.fatal(msg.error);

                // TODO: if publish failed, look why and give error accordingly
                    // if b/c profile clash/app name clash => prompt for new name
            })
            .then(function(msg) {
                console.log(msg);
            })
            .fail(function(msg) {
                console.log('got to fail');
                self.helpers.fatal(msg);
            });

    },

    // NOT IN USE
    removeAppRemote: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>

        // TODO
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Create app instance
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);
        var name = this.app.getName();

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Make sure all the required attributes are present before publishing
        var validApp = this.controllers.app.validateApp();
        if (!validApp.valid) {
            this.helpers.fatal(validApp.msg);
//            this.helpers.fatal('Missing required fields in package.json. Note: both name and version are required.');

            // TODO: prompt for missing required fields?
        }

        var self = this;

        this.helpers.removeAppRemote(credentials, name)
            .then(function() {

                console.log('Successfully removed ' + name + ' from the server.');

            }, function(err) {

                self.helpers.fatal('Failed to remove ' + name + ' from the server.');

            });
    },

    /*
    changeProfileName: function(argv) {

        var profileName = argv._.shift();

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        // Create Profile entry and save it
        var profile = this.controllers.app.changeProfileName(profileName);
        this.app.save();

        // TODO: make pretty
        if (profile.success) {
            console.log(JSON.stringify(profile.profile, null, 4));
        } else {
            this.helpers.fatal(profile.message);
        }

    },
    */

    // Note this is local
    viewProfile: function(argv) {

        // Create app instance
        // TODO: abstract this out into a fn!!
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.viewProfile();

        if (profile.success) {
            console.log(JSON.stringify(profile.profile, null, 4));
        } else {
            this.helpers.fatal(profile.message);
        }

    },

    addTypeToProfile: function(argv) {

        var typeName = argv._.shift(),
            actions = {
                select: argv.s,
                insert: argv.i,
                deleteAction: argv.d,
                aggregate: argv.a,
                update: argv.U
            };

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.addTypeToProfile(typeName, actions);
        this.app.save();

        if (profile.success) {
            console.log(JSON.stringify(profile.profile, null, 4));
        } else {
            this.helpers.fatal(profile.message);
        }

    },

    removeTypeFromProfile: function(argv) {
        var typeName = argv._.shift();

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        var profile = this.controllers.app.removeTypeFromProfile(typeName);

        this.app.save();

        if (profile.success) {
            console.log(JSON.stringify(profile.profile, null, 4));
        } else {
            this.helpers.fatal(profile.message);
        }
    },

    assignProfileToUser: function(argv) {

        var username = argv._.shift();

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Create app instance
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        this.helpers.assignProfile(this.app.getProfile().name, username, credentials)
            .then(function(msg) {
                console.log(msg);
            });
    },

    // NOT IN USE
    addProfile: function(argv) {

        // ENDPOINT: POST /apps/<app name>/profiles/<profile name (KFC.user)>

        this.profileCurrier('add')(argv);

    },

    // NOT IN USE
    removeProfile: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>/<profile name>

        this.profileCurrier('remove')(argv);
    },

    listProfile: function(argv) {

        // ENDPOINT: GET /apps/<app name>

        this.profileCurrier('list')(argv);
    },

    profileCurrier: function(type) {

        var self = this,
            successMessage = '',
            failMessage = '',
            fn;

        switch(type) {
            case 'add':
                fn = this.helpers.addProfile;
                successMessage = 'Successfully added profile';
                failMessage = 'Failed to add Profile.';
                break;
            case 'remove':
                fn = this.helpers.removeProfile;
                successMessage = 'Successfully removed profile';
                failMessage = 'Failed to remove Profile.';
                break;
            case 'list':
                fn = this.helpers.listProfile;
                failMessage = 'Failed to get profile.';
                break;
        }

        return function(argv) {

            var profileName;

            if (type === 'remove' || type === 'add') {
                profileName = argv._.shift();
                if (!profileName) {
                    self.helpers.fatal('Missing profile name. Please specify one.');
                }
            }

            self.demandProjectOrExit();

            // Validate credentials
            var credentials = self.helpers.validateCredentials(self.project, argv);

            var pathToAppFolder = self.helpers.getPathToAppFolder(self.cwd);
            self.app = self.helpers.loadPackageSync(pathToAppFolder);

            fn(self.app.getName(), credentials, profileName)
                .then(function(profile) {

                    if (profile) {
                        console.log(profile);

                        // TODO: show dml of profile!
                    } else {
                        console.log(successMessage);
                    }

                }, function(err) {
                    self.helpers.fatal(err);
                });
        }
    },

    listApps: function(argv) {

        // ENDPOINT: Get /apps

        var credentials = this.helpers.validateCredentials(this.project, argv),
            self = this;

        this.helpers.listApps(credentials)
            .then(function(apps) {
                console.log(apps);
            }, function() {
                self.helpers.fatal('Failed to get apps. Please try again.');
            });
    },

    // NOTE: these are repeats... abstract this out for all the commands?
    demandProjectOrExit: function(){
        if(!this.project){
            this.helpers.fatal('Unknown project. Please init a new project in this directory or move to an active project.');
        }
    }
};

// MAIN
new AppTools(argv).dispatch(argv);