module.exports = function(project, options){

    var _  = require('../utils.js')
      , path    = require('path')
      , fs      = require('fs-extra')
      , format  = require('format-json-stream')
      , Resource= require('../types').Resource
      , Embedded= require('../types').Embedded
      , Property= require('../types').Property
    ;

    /**
     * Create the model
     */
    function create(typeName, spec){

        if(options.verbose) console.log('Creating a new type ' + name + ' ...');

        //map
        var name        = typeName;
        var ns          = spec.namespace;
        var storeAs     = spec.storage_name || name;
        var type;

        //Belongs in a processor
        if(spec.embedded){
            type = new Embedded(typeName);
        } else {
            type = new Resource(typeName);
            type.storedAs(storeAs);
            if(spec.immutable) type.asImmutable();
            if(spec.extensible) type.asExtensible();
            if(spec.encapsulated) type.asEncapsulated();
        }

        save(type);
    }


    function rename(from, to){

        if(options.verbose) console.log('Renaming the type ' + from + ' to ' + to + '...');

        //test
        if(from == null){
            console.log('Unspecified Name.')
        }

        if(to == null){
            console.log('Unspecified new Name.')
        }

        //Belongs in the processor
        var type = load(asFilePath(from));

        type.name = to;

        if(type.storage_name === from){
            type.storage_name = to;
        }

        save(type);
        remove(from);

        console.log(type.name);
    }

    /**
     * Show a local model
     *
     * -color    : print in color
     * -minimal  : only show the required fields
     * -flat     : do not indent
     *
     */
    function view(typeName, view){
        if(options.verbose) console.log('Showing the local type ' + typeName + '...');
        try{
            var model = fs.readJsonFileSync(asFilePath(typeName));
            view.render(model);
        } catch (err){
            console.log('No such model %s', typeName)
        }
    }

    /**
     * Get the model from the server
     */
    function edit(typeName, patch){

        if(options.verbose) console.log('Editing the local type ' + typeName + '...');

        //Push Down to the type models
        var type = load(asFilePath(typeName));

        //Reset the namespace
        if(patch.namespace)    type.inNamespace(patch.namespace);

        //-Converting to embedded
        if(patch.embedded){
            type.embedded = true;
            delete type.storage_name;
            delete type.immutable;
            delete type.encapsulated;
            delete type.extensible;
        } else {
            if(patch.storage_name)  type.storage_name = patch.storage_name;
            if(patch.immutable)    type.storage_name = patch.immutable;
            if(patch.encapsulated) type.storage_name = patch.encapsulated;
            if(patch.extensible)   type.storage_name = patch.extensible;
        }

        save(type);
    }

    /**
     * Delete locally
     */
    function asFilePath(name){
        return path.join(process.cwd(), 'models', (name + '.json'));
    }

    /**
     * Load a Model
     * @param modelName
     * @returns {Embedded}
     */
    function load(typeName){
        try{
            var raw = fs.readJsonSync(typeName);
            var type = (raw.embedded) ? new Embedded() : new Resource();
            type.merge(raw);
            return  type;
        } catch (err) {
            return  undefined;
        }
    }

    function save(resource){
        var filePath = asFilePath(resource.name);

        fs.outputFile(filePath, _.toDelimitedJson(resource), function(err) {
            if(err){
                if(options.verbose) console.log('Unable to write the project file');
            } else {
                console.log(filePath);
            }
        });
    }

    /**
     * Remove the model file
     * @param name
     */
    function remove(typeName){

        if(options.verbose) console.log('Removing the local type ' + typeName + '...');

        fs.remove(asFilePath(typeName), function(err){
            if (err) return console.error("Error: Unable to delete the Model " + typeName);
            if(options.verbose) console.log('Type ' + typeName + ' Deleted');
        });
    }


    function getProps(typeName){
        var type = load(asFilePath(typeName));
        _.forEach(type.properties, function(value, key, list){
            if(!_.startsWith(key, 'rbc_')){
                console.log(key);
            }
        })
    }

    function genInstance(typeName){
        var type = load(asFilePath(typeName));

        //get the property names
        var propertyNames = _.keys(type.properties);
        propertyNames.sort();

        var instance = {};

        for(var i = 0; i < propertyNames.length; i++){
            var key  = propertyNames[i];
            var prop = type.properties[key];

            //For all non rbc attributes
            if(!_.startsWith(key, 'rbc_')){

                var pType = prop.type.toLowerCase();
                var pValue=null;
                switch(pType){
                    case 'boolean':
                        pValue = false;
                        break;
                    case 'integer':
                        pValue = 0;
                        break;
                    case 'real':
                        pValue = 0.00;
                        break;
                    case 'string':
                        pValue = "";
                        break;
                    case 'datetime':
                        pValue = new Date().toISOString();
                        break;
                    case 'decimal':
                        pValue = "0.00";
                        break;
                    case 'currency':
                        pValue =  "USD:0.00";
                        break;
                    case 'geojson':
                        pValue = [0.00, 0.00];
                        break;
                    default:
                        pValue = {};
                        break;
                }
                instance[key] = (prop.multi) ? [pValue]: pValue;
            }
        }
        console.log(_.toDelimitedJson(instance));
    }

    function putProp(typeName, properties, spec){

        var type = load(asFilePath(typeName));

        _.forEach(properties, function(prop_name){

            var prop = new Property(prop_name)

            //Select the type
            if(spec.type){
                prop.ofType(spec.type)
            } else if (spec.Boolean){
                prop.asBoolean();
            } else if (spec.Integer){
                prop.asInteger();
            } else if (spec.Real){
                prop.asReal();
            } else if (spec.String){
                prop.asString();
            } else if (spec.Link){
                prop.asLink();
            } else if (spec.DateTime){
                prop.asDateTime();
            } else if (spec.Decimal){
                prop.asDecimal();
            } else if (spec.Currency){
                prop.asCurrency();
            } else if (spec.GeoJSON){
                prop.asGeoJSON();
            } else {
                prop.asObject();
            }

            //Property Options
            if(spec.required)   prop.asRequired(true);
            if(spec.unique)     prop.asUnique(true);
            if(spec.multi)      prop.asList(true);
            if(spec['default']) prop.withDefault(spec['default'])
            if(spec.scale)      prop.ofScale(spec.scale);


            type.bindProperty(prop);
        })

        save(type);
    }

    function mvProp(typename, from, to){

        var type = load(asFilePath(type_name));

        type.properties = type.properties || {};

        if(type.properties[from_name]){
            type.properties[to_name] = type.properties[from_name];
            delete type.properties[from_name];
        }

        save(type);
    }

    function rmProp(typeName, property){
        var type = load(asFilePath(type_name));
        delete type.properties[prop_name];
        save(type);
    }

    return {
        create : create,
        rename : rename,
        view   : view,
        edit   : edit,
        remove : remove,
        getProps: getProps,
        putProp: putProp,
        mvProp : mvProp,
        rmProp : rmProp,
        genInstance: genInstance
    }

}

